<div class="message message-ai">
<p>
    Sigmund: Welcome stranger! 🤗 I am Sigmund, an AI research assistant powered by the most powerful technologies currently available.
</p>

<p>
    Like other chatbots, I can help you with pretty much anything. However, a few unique features set me apart from other chatbots, such as ChatGPT, Claude, or DeepSeek. These features make me the perfect scientific side kick, especially if you're a cognitive neuroscientist, psychologist, or data scientist!
</p>
    
<ul>
    <li>
        I have expert knowledge of OpenSesame and other scientific Python packages that other chatbots struggle with 👩🏽‍🔬
    </li>
    <li>
        I can search Google Scholar to find research papers and verify information in primary sources 📚  
    </li>
    <li>
        I store all your data in an encrypted database that no-one can read, not even my administrators! 🔒
    </li>
    <li>
        I am able to understand and generate images 🖼️
    </li>
</ul>

<p>
    To get started, <a onclick="app.__vue__.toggleMenu()" href=".">sign in</a> with your Google account or your account from forum.cogsci.nl. Once you're signed in, you need to subscribe (if you're not subscribed already!).
</p>
<p>
    A subscription is just €9 per month, which is much cheaper than other premium chatbots! Subscribing is easy and secure, with a variety of international payment methods available. If you need a custom subscription, such as an invoiced subscription for multiple individuals or a longer period, contact <a href="mailto:professional@cogsci.nl">professional@cogsci.nl</a>.
</p>
<p>I hope to talk to you soon! 💖</p> 
</div>